from libc.stdlib cimport malloc, free
from libc.string cimport strdup


# Declare main_fasta function from C
cdef extern from "lib/include/km_fasta.h":
    int main_fasta(int argc, char **argv)

def py_main_fasta(list argv):
    cdef int argc = len(argv)
    
    if argc == 0:
        # Handle empty argv gracefully
        return
    
    cdef char **c_argv = <char **>malloc((argc+1) * sizeof(char *))
    if c_argv is NULL:
        # Handle allocation failure
        raise MemoryError("Failed to allocate memory")
    
    try:
        for i in range(argc):
            if argv[i] is None:
                # Handle invalid argument
                raise ValueError("Invalid argument at index {}".format(i))
            c_argv[i] = strdup(argv[i].encode('utf-8'))
            if c_argv[i] is NULL:
                # Handle strdup failure
                raise MemoryError("Failed to allocate memory for argument {}".format(i))
        c_argv[argc] = NULL
        return main_fasta(argc, c_argv)
    finally:
        for i in range(argc):
            if c_argv[i] is not NULL:
                free(c_argv[i])
        free(c_argv)



from re import M
import sys
from xmlrpc.client import boolean
import subprocess
from datetime import datetime
from colorama import Fore, Style


def eprint(*args, **kwargs):
    log = open('kmat_tools.log', 'a')
    log.write(*args)
    log.write('\n')
    print(*args, file=sys.stderr, **kwargs)

def datetime_now():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

def print_error(msg):
    eprint(f'[{datetime_now()}] [' + f'{Fore.RED}error{Style.RESET_ALL}' + f'] {msg}')

def print_warning(msg):
    eprint(f'[{datetime_now()}] [' + f'{Fore.YELLOW}warning{Style.RESET_ALL}' + f'] {msg}')

def print_status(msg):
    eprint(f'[{datetime_now()}] [' + f'{Fore.GREEN}status{Style.RESET_ALL}' + f'] {msg}')

def remove_files(path):
    subprocess.run(["rm", "-rf", path],capture_output=True,text=True)




import argparse
import km_fasta

def parse_arguments():
    parser = argparse.ArgumentParser(description="Outputs k-mers of a k-mer matrix in a FASTA file.")

    parser.add_argument("in_mat", metavar="<in.mat>", help="Input file containing k-mer matrix")
    parser.add_argument("-o",metavar="FILE", help="Output FASTA file of k-mers to FILE (default: stdout)")
    return parser

def km_fasta_main():
    args_km_fasta=["fasta"]
    try:
        parser=parse_arguments()
        args=parser.parse_args()
        for name, value in vars(args).items():
            if value!=None and name!="in_mat":
                args_km_fasta=args_km_fasta+["-"+name,str(value)]
        args_km_fasta=args_km_fasta+[args.in_mat]
        km_fasta.py_main_fasta(args_km_fasta)
        
        return 0

    except Exception as e:
        print(f"Error: {str(e)}")
        return 1

if __name__ == "__main__":
    km_fasta_main()



import argparse
import km_unitig  

def parse_arguments():
    parser = argparse.ArgumentParser(description="Creates a unitig matrix.")

    parser.add_argument("unitigs_file", metavar="<unitigs.fasta>", help="Input file containing unitigs in FASTA format")
    parser.add_argument("kmer_matrix", metavar="<kmer_matrix>", help="Input file containing kmer matrix")
    parser.add_argument("-k", type=int, default=31, help="Size of k-mers of input matrices (default: 31)")
    parser.add_argument("-l",  type=int, default=0, help="Minimum length of unitigs to consider (default: 0)")
    parser.add_argument("-o", metavar="FILE", help="Write unitig matrix to FILE (default: stdout)")
    
    return parser

def km_unitig_main():
    args_km_unitig=["filter"]
    try:
        parser=parse_arguments()
        args=parser.parse_args()
        for name, value in vars(args).items():
            if value!=None and name!="unitigs_file" and name!="kmer_matrix":
                args_km_unitig=args_km_unitig+["-"+name,str(value)]
        args_km_unitig=args_km_unitig+[args.unitigs_file,args.kmer_matrix]
        km_unitig.py_main_unitig(args_km_unitig)
        
        return 0

    except Exception as e:
        print(f"Error: {str(e)}")
        return 1

if __name__ == "__main__":
    km_unitig_main()



#!/usr/bin/env python3
import os
import logging
import sys
import subprocess
import shutil
import argparse
from kmat_tools.utils.utils import eprint,print_error, print_warning, print_status
import km_reset
from kmat_tools.__version__ import __version__
# Default parameters
DEFAULT_KSIZE = 31
DEFAULT_THREADS = 4
DEFAULT_UTG_MIN_SIZE = 100
DEFAULT_MIN_COUNT = 1
DEFAULT_KMAT_ABUNDANCE = 1
DEFAULT_KMAT_MIN_ZEROS = 0
DEFAULT_KMAT_MIN_NONZEROS = 0
DEFAULT_MIN_REC = 3
DEFAULT_OUTDIR = "output"

kmat_tools_root = os.path.dirname(os.path.abspath(os.path.realpath(os.__file__)))


def print_version():
    kmat_tools_version = f'{__version__}'
    git = subprocess.run(['git', '-C', kmat_tools_root, 'describe', '--always'], capture_output=True, stderr=None, text=True)
    if git.returncode == 0:
        git_hash = git.stdout.strip().rsplit('-', 1)[-1]
        kmat_tools_version += f'-{git_hash}'
    return kmat_tools_version

import subprocess
import os

def replace_unitig_identifier(output_dir):
    # Command to sort unitigs.mat file
    sort_command = f"sort -n -o {output_dir}/unitigs.mat {output_dir}/unitigs.mat"

    # Command to extract unitig sequences
    awk_commands = f"""
    awk '{{print $1}}' {output_dir}/unitigs.mat | while read line; do
        awk -v term=">$line" '$1 == term {{getline; print; exit}}' {output_dir}/unitigs.fa
    done > {output_dir}/unitigs_tmp.fa
    """

    # Command to paste and cut
    cut_paste_command = f"""
    cut -f 2- -d ' ' {output_dir}/unitigs.mat |
    paste {output_dir}/unitigs_tmp.fa - -d " " > {output_dir}/unitigs_tmp.mat
    """

    # Execute commands
    subprocess.run(sort_command, shell=True, check=True)
    subprocess.run(awk_commands, shell=True, check=True)
    subprocess.run(cut_paste_command, shell=True, check=True)

    # Clean up
    subprocess.run(f"rm {output_dir}/unitigs.mat {output_dir}/unitigs_tmp.fa", shell=True, check=True)
    subprocess.run(f"mv {output_dir}/unitigs_tmp.mat {output_dir}/unitigs.mat", shell=True, check=True)
    

def main():
    parser = argparse.ArgumentParser(description='Create unitig matrix from fof.txt file and set of FASTA/FASTQ files using kmat_tools')
    parser.add_argument('input_file', metavar='input_seqfile', type=str, help='input seqfile (fof)')
    parser.add_argument('-k', '--kmer-size', type=int, default=DEFAULT_KSIZE, help='kmer size for ggcat and kmtricks (default: 31)')
    parser.add_argument('-t', '--threads', type=int, default=DEFAULT_THREADS, help='number of cores (default: 4)')
    parser.add_argument('-u', '--min-utg-size', type=int, default=DEFAULT_UTG_MIN_SIZE, help='minimum size of the unitigs to be retained in the final matrix (default: 100)')
    parser.add_argument('-c', '--min-count', type=int, default=DEFAULT_MIN_COUNT, help='minimum count of kmers to be retained (default: 1)')
    parser.add_argument('-o', '--outdir', type=str, default=DEFAULT_OUTDIR, help='output directory (default: output)')
    parser.add_argument('-r', '--min-rec', type=int, default=DEFAULT_MIN_REC, help='minimum recurrence to keep a k-mer (default: 3)')
    parser.add_argument('-a', '--min-abundance', type=int, default=DEFAULT_KMAT_ABUNDANCE, help='minimum abundance for kmtool (default: 1)')
    parser.add_argument('-n', '--min-zeros', type=int, default=DEFAULT_KMAT_MIN_ZEROS, help='minimum number of zero columns for each kmers (for kmtools) (default: 0)')
    parser.add_argument('-N', '--min-nonzeros', type=int, default=DEFAULT_KMAT_MIN_NONZEROS, help='minimum number of non-zero columns for each kmers (for kmtools) (default: 0)')
    parser.add_argument('-V', '--version', action='version', version=f'kmat_tools {__version__}',help='Show version number and exit')
    
    args = parser.parse_args()

    if os.path.exists("kmat_tools.log"):
        os.remove("kmat_tools.log")

    logging.basicConfig(filename='kmat_tools.log', level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(message)s')
    #sys.stdout = sys.stderr = open('kmat_tools.log', 'a')

    print_status("Starting kmat_tools version: " + __version__)


    # Parameters
    print_status("kmer size:" + str(args.kmer_size))
    print_status("Number of threads:" + str(args.threads))
    print_status("Minimum unitig size:"+ str(args.min_utg_size))
    print_status("Minimum kmer count:"+ str(args.min_count))
    print_status("Output directory:"+ str(args.outdir))
    print_status("Minimum recurrence:"+ str(args.min_rec))
    print_status("Additional options:")
    print_status("Minimum abundance:"+ str(args.min_abundance))
    print_status("Minimum number of zero columns:"+ str(args.min_zeros))
    print_status("Minimum number of non-zero columns:"+ str(args.min_nonzeros))
    print_status("Input file:"+ args.input_file)


    # Building output directory if not present already
    if os.path.isdir(str(args.outdir)):
        print_warning("Folder "+ args.outdir + " is going to be overwritten.")
        shutil.rmtree(str(args.outdir))

    # Step 1: Produce sorted input k-mer matrix from fof.txt file
    print_status("[PIPELINE]::[running kmtricks]")

    try:
        pipeline = subprocess.Popen(["kmtricks", "pipeline", "--file", args.input_file, "--kmer-size", str(args.kmer_size), "--hard-min",
                    str(args.min_count), "--mode", "kmer:count:bin", "--cpr", "--run-dir", args.outdir, "-t", str(args.threads),
                    "--recurrence-min", str(args.min_rec)],stdout=subprocess.PIPE, text=True,stderr=subprocess.STDOUT)
        for line in pipeline.stdout:
            if line != "\n":
                line = line.rstrip('\n')  # Strip the final newline character
                print_status(line)
        
        pipeline.wait()

    except Exception as e:
        print_error(e)
        return 1

    try:
        aggregate = subprocess.Popen(["kmtricks", "aggregate", "--matrix", "kmer", "--format", "text", "--cpr-in", "--sorted", "--output",
                    os.path.join(args.outdir, "sorted_matrix.txt"), "--run-dir", args.outdir, "-t", str(args.threads)],\
                        stdout=subprocess.PIPE, text=True,stderr=subprocess.STDOUT)
        for line in aggregate.stdout:
            if line != "\n":
                line = line.rstrip('\n')  # Strip the final newline character
                print_status(line)
        aggregate.wait()

    except Exception as e:
        print_error(e)
        return 1
    
    # Step 2: Perform some basic filtering with km_basic_filter
    print_status("[PIPELINE]::[filtering kmers]")
    km_reset.reset_getopt()
    try:
    # Run km_basic_filter.py_main_basic_filter as a subprocess
        command_km_basic_filter = ["km_basic_filter", "-a",
                str(args.min_abundance), "-n", str(args.min_zeros), "-N", str(args.min_nonzeros), "-o",
                os.path.join(args.outdir, "sorted_filtered_matrix.txt"),
                os.path.join(args.outdir, "sorted_matrix.txt")]
        
        # Open the log file for appending
        with open('kmat_tools.log', 'a') as log_file:
            process = subprocess.Popen(command_km_basic_filter, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

            # Capture the output and write it to the log file
            for line in process.stdout:
                line = line.rstrip('\n')  # Strip the final newline character
                print_status(line)

            # Wait for the subprocess to finish
            process.wait()
        if process.returncode != 0:
            print_error("Error in km_basic_filter.py_main_basic_filter")
            return(1)
    except Exception as e:
        print_error(e)
        print_error("Error in km_basic_filter.py_main_basic_filter")
        return(1)

    # Step 3: Build unitigs
    # Step 3.1: Convert sorted filtered matrix into a fasta file
    print_status("[PIPELINE]::[exporting filtered kmers]")
    #Reset getopts global variable
    km_reset.reset_getopt()
    try:
        command_km_fasta=["km_fasta", "-o", os.path.join(args.outdir, "kmer_matrix.fasta"),
                                os.path.join(args.outdir, "sorted_filtered_matrix.txt")]
        process = subprocess.Popen(command_km_fasta,
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    
        # Capture the output and write it to the log file
        for line in process.stdout:
            line = line.rstrip('\n')  # Strip the final newline character
            print_status(line)

        # Wait for the subprocess to finish
        process.wait()

        if process.returncode != 0:
            print_error("Error in km_fasta.py_main_fasta")
            return 1
    except Exception as e:
        print_error(e)
        print_error("Error in km_fasta.py_main_fasta")
        return 1
    
    # Step 3.2: Use ggcat to produce unitigs
    print_status("[PIPELINE]::[generating unitigs]")
    try:
        command_ggcat=["ggcat", "build", os.path.join(args.outdir, "kmer_matrix.fasta"), "-o",
                    os.path.join(args.outdir, "unitigs.fa"), "-j", str(args.threads), "-s", str(args.min_count), "-k",
                    str(args.kmer_size)]
        process=subprocess.Popen(command_ggcat,stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

            # Capture the output and write it to the log file
        for line in process.stdout:
            line = line.rstrip('\n')  # Strip the final newline character
            print_status(line)

        # Wait for the subprocess to finish
        process.wait()
    
    except Exception as e:
        print_error(e)
        print_error("Error in ggcat")
        return(1)

    # Step 4: Build unitig matrix
    print_status("[PIPELINE]::[building unitigs matrix]")
    km_reset.reset_getopt()
    command_km_unitig=["km_unitig","-l", str(args.min_utg_size), "-k", str(args.kmer_size), "-o",
                   os.path.join(args.outdir, "unitigs.mat"),os.path.join(args.outdir, "unitigs.fa"),
                   os.path.join(args.outdir, "sorted_filtered_matrix.txt")]
    try:
        # Open the log file for appending
        with open('kmat_tools.log', 'a') as log_file:
            process = subprocess.Popen(command_km_unitig, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

            # Capture the output and write it to the log file
            for line in process.stdout:
                line = line.rstrip('\n')  # Strip the final newline character
                print_status(line)

            # Wait for the subprocess to finish
            process.wait()
        if process.returncode != 0:
            print_error("Error in km_unitig")
            return(1)
    except Exception as e:
        print_error(e)
        print_error("Error in km_unitig")
        return(1) 

    # Step 5: Replace unitig identifier in unitigs.mat by unitig sequence
    print_status("[PIPELINE]::[replacing unitig identifier]")
    replace_unitig_identifier(args.outdir)  

    print_status("kmat_tools finished successfully")

if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        eprint(e)
        print_error(e)  




# km_select.pyx
from libc.stdlib cimport abort, calloc, free
from libc.stdio cimport fopen, fclose, fputs, stderr, stdout
from libc.string cimport strcmp
from cpython.mem cimport PyMem_Malloc, PyMem_Free

cdef extern from "common.h":
    bint next_kmer(char *, int, FILE *)
    bint next_kmer_and_line(char *, int, char **, size_t *, FILE *)
    int ktcmp(const char *, const char *)

def main_select(int ksize=31, char *out_fname=None, bint do_select=True, bint use_ktcmp=False):
    cdef:
        FILE *selfile
        FILE *matfile
        FILE *outfile
        char *sel_kmer = <char *>calloc(ksize+1,1)
        char *mat_kmer = <char *>calloc(ksize+1,1)
        char *line = NULL
        size_t line_size = 0
        bint ret_sel, ret_mat
        size_t tot_kmers, kept_kmers = 0
        int ret_cmp

    selfile = fopen("matrix_1","r")
    if selfile == NULL:
        print(f"Cannot open file matrix_1")
        return 1

    matfile = fopen("matrix_2","r")
    if matfile == NULL:
        print(f"Cannot open file matrix_2")
        fclose(selfile)
        return 1

    outfile = fopen(out_fname,"w") if out_fname else stdout
    if outfile != stdout and outfile == NULL:
        print(f"Cannot open output file {out_fname}")
        fclose(selfile)
        fclose(matfile)
        return 1

    ret_sel = next_kmer(sel_kmer, ksize, selfile)
    ret_mat = next_kmer_and_line(mat_kmer, ksize, &line, &line_size, matfile)
    tot_kmers = ret_mat
    while ret_sel and ret_mat:
        ret_cmp = ktcmp(sel_kmer,mat_kmer) if use_ktcmp else strcmp(sel_kmer, mat_kmer)
        if ret_cmp == 0:
            if do_select:
                fputs(line,outfile)
                kept_kmers += 1
            ret_sel = next_kmer(sel_kmer, ksize, selfile)
            ret_mat = next_kmer_and_line(mat_kmer, ksize, &line, &line_size, matfile)
            tot_kmers += ret_mat
        elif ret_cmp < 0:
            ret_sel = next_kmer(sel_kmer, ksize, selfile)
        else: # ret_cmp > 0
            if not do_select:
                fputs(line,outfile)
                kept_kmers += 1
            ret_mat = next_kmer_and_line(mat_kmer, ksize, &line, &line_size, matfile)
            tot_kmers += ret_mat

    while ret_mat:
        if not do_select:
            fputs(line,outfile)
            kept_kmers += 1
        ret_mat = next_kmer_and_line(mat_kmer, ksize, &line, &line_size, matfile)
        tot_kmers += ret_mat

    print(f"[info] {tot_kmers}\ttotal k-mers")
    print(f"[info] {kept_kmers}\tretained k-mers")

    free(line)
    fclose(selfile)
    fclose(matfile)
    if outfile != stdout:
        fclose(outfile)

    return 0


import argparse
import km_basic_filter  

def parse_arguments():
    parser = argparse.ArgumentParser(description="Filter a matrix by selecting k-mers that are potentially differential.")
    
    parser.add_argument("in_mat", metavar="<in.mat>", type=str, help="Input matrix file")
    parser.add_argument("-a", type=int, default=10, help="Minimum abundance to define a k-mer as present in a sample [10]")
    parser.add_argument("-n", type=int, default=10, help="Minimum number of samples for which a k-mer should be absent [10]")
    parser.add_argument("-f", type=float, help="Fraction of samples for which a k-mer should be absent (overrides -n)")
    parser.add_argument("-N", type=int, default=10, help="Minimum number of samples for which a k-mer should be present [10]")
    parser.add_argument("-F", type=float, help="Fraction of samples for which a k-mer should be present (overrides -N)")
    parser.add_argument("-o", type=str, help="Output filtered matrix to FILE [stdout]")
    
    return parser

def km_basic_filter_main():
    args_km_basic_filter=["filter"]
    try:
        parser=parse_arguments()
        args=parser.parse_args()
        for name, value in vars(args).items():
            if value!=None and name!="in_mat":
                args_km_basic_filter=args_km_basic_filter+["-"+name,str(value)]
        args_km_basic_filter=args_km_basic_filter+[args.in_mat]
        km_basic_filter.py_main_basic_filter(args_km_basic_filter)
        
        return 0

    except Exception as e:
        print(f"Error: {str(e)}")
        return 1

if __name__ == "__main__":
    km_basic_filter_main()



from setuptools import setup,Extension,find_packages
from Cython.Build import cythonize

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

EXTENSION_PATH = "cython_helpers"

ext1 = Extension("km_basic_filter",
                 sources=[f"{EXTENSION_PATH}/km_basic_filter.pyx", f"{EXTENSION_PATH}/lib/src/km_basic_filter.c"],
                 include_dirs=[f"{EXTENSION_PATH}/lib/include"],
                 extra_compile_args=['-Wall', '-Wno-unused-function', '-O3', '-pedantic'],
                 libraries=['z', 'm'])

ext2 = Extension("km_fasta",
                 sources=[f"{EXTENSION_PATH}/km_fasta.pyx", f"{EXTENSION_PATH}/lib/src/km_fasta.c"],
                 include_dirs=[f"{EXTENSION_PATH}/lib/include"],
                 extra_compile_args=['-Wall', '-Wno-unused-function', '-O3', '-pedantic'],
                 libraries=['z', 'm'])

ext3 = Extension("km_unitig",
                 sources=[f"{EXTENSION_PATH}/km_unitig.pyx", f"{EXTENSION_PATH}/lib/src/km_unitig.c"],
                 include_dirs=[f"{EXTENSION_PATH}/lib/include"],
                 extra_compile_args=['-Wall', '-Wno-unused-function', '-O3', '-pedantic'],
                 libraries=['z', 'm'])

ext4 = Extension("km_reset",
                 sources=[f"{EXTENSION_PATH}/km_reset.pyx", f"{EXTENSION_PATH}/lib/src/km_reset.c"],
                 include_dirs=[f"{EXTENSION_PATH}/lib/include"],
                 extra_compile_args=['-Wall', '-Wno-unused-function', '-O3', '-pedantic'],
                 libraries=['z', 'm'])

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="kmat_tools",
    version="0.0.17",
    author="Riccardo Vicedomini, Francesco Andreace, Camila Duitama",
    author_email="cduitama@pasteur.fr",
    include_package_data=True,
    packages=find_packages(include=["kmat_tools", "utils", "cython_helpers", "cython_helpers.km_tools_wrappers"]),
    package_data={"": ["kmat_tools/*",
                       "cython_helpers/*.pxd",
                       "cython_helpers/*.pyx",
                       "cython_helpers/lib/include/*.h",
                       "cython_helpers/lib/src/*.c",
                       "cython_helpers/km_tools_wrappers/*.py"]},
    ext_modules = cythonize([ext1,ext2,ext3,ext4]),
    entry_points={
        "console_scripts": [
            "kmat_tools = kmat_tools.__main__:main",
            'km_basic_filter = cython_helpers.km_tools_wrappers.km_basic_filter_wrapper:km_basic_filter_main',
            'km_fasta = cython_helpers.km_tools_wrappers.km_fasta_wrapper:km_fasta_main',
            'km_unitig = cython_helpers.km_tools_wrappers.km_unitig_wrapper:km_unitig_main'
       ]
    },
        install_requires=[
        "colorama",
        "Cython",
        "wheel",
        "setuptools",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CamilaDuitama/kmat_tools",
    project_urls={
        "Bug Tracker": "https://github.com/CamilaDuitama/kmat_tools/issues",
    },
        classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
